#charset "us-ascii"

/*
 *   Copyright 2003-04 Soren J. Lovborg.
 *  
 *   Permission is hereby granted, free of charge, to any person obtaining a
 *   copy of this software and associated documentation files (the "Software"), 
 *   to deal in the Software without restriction, including without limitation 
 *   the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *   and/or sell copies of the Software, and to permit persons to whom the
 *   Software is furnished to do so, subject to the following conditions:
 *  
 *   The above copyright notice and this permission notice shall be included in
 *   all copies or substantial portions of the Software.
 *  
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

#include <tads.h>
#define false nil

/* ------------------------------------------------------------------------ */

enum token tokString, tokNumber, tokOperator, tokWord, tokEmit, tokEOF;

Token template 'token';

class Token: object
    construct(value_) { value = value_; }
    className = nil
    token = nil

    getCompilerValue()
    {
        throw new ECompilerError('Tokens of type ' + className + ' have no value');
    }
;
class ValueToken: Token; // Tokens which have a simple value.
class WordToken: Token;  // Tokens which match a word / valid identifier.

class EmitToken: Token;
class JCFunctionToken: WordToken  construct(tok_) { token = tok_; };

class NumberToken: ValueToken  getCompilerValue = (new CompilerConstantNumber(value));
class StringToken: ValueToken  getCompilerValue = (CompilerConstant.createFromString(value));
class IdentifierToken: WordToken, ValueToken
    getCompilerValue = (new CompilerIdentifier(self))

    identifierValue_ = nil
    identifierValue()
    {
        if (!identifierValue_) throw new ECompilerError('No value assigned to identifier "' + value + '"');
        return identifierValue_;
    }
    assignValue(val)
    {
        if (identifierValue_) throw new ECompilerError('Identifier "' + value + '" has already been assigned a value');
        identifierValue_ = val;
    }
;

eofToken: Token;

tPeriod:        Token '.';
tColon:         Token ':';
tSemicolon:     Token ';';
tParenBegin:    Token '(';
tParenEnd:      Token ')';
tBraceBegin:    Token '{';
tBraceEnd:      Token '}';
tBracketBegin:  Token '[';
tBracketEnd:    Token ']';
tComma:         Token ',';
tAssign:        Token '=';
tConditional:   Token '?';

rwConst:        WordToken 'const';
rwCase:         WordToken 'case';
rwDefault:      WordToken 'default';
rwSwitch:       WordToken 'switch';
rwIf:           WordToken 'if';
rwThen:         WordToken 'then';
rwElse:         WordToken 'else';
rwLet:          WordToken 'let';
rwContentsloop: WordToken 'contentsloop';
rwFieldloop:    WordToken 'fieldloop';
rwStrloop:      WordToken 'strloop';
rwRepeat:       WordToken 'repeat';
rwBy:           WordToken 'by';
rwIn:           WordToken 'in';
rwIs:           WordToken 'is';
rwIsnt:         WordToken 'isnt';
rwOn:           WordToken 'on';
rwSay:          WordToken 'say';
rwOSay:         WordToken 'osay';
rwRSay:         WordToken 'rsay';
rwArrayLen:     WordToken 'arraylen';

rwThis:         WordToken, ValueToken 'this'
    getCompilerValue()
    {
        if (jccCompiler.currentlyDefiningObject)
            return new CompilerConstantNumber(jccCompiler.currentlyDefiningObject.id);
        throw new ECompilerError('Reference to "this" outside object definition');
    }
;

rwVarObj:       WordToken, ValueToken 'varobj'
    getCompilerValue = (new CompilerVarObjReference())
;

class ObjectTypeToken: WordToken;

rwRoom:   ObjectTypeToken                  'room'   type = 1;
rwPlayer: JCFunctionToken, ObjectTypeToken 'player' type = 2;
rwAction: ObjectTypeToken                  'action' type = 3;
rwExit:   JCFunctionToken, ObjectTypeToken 'exit'   type = 3;
rwThing:  ObjectTypeToken                  'thing'  type = 4;

/* ------------------------------------------------------------------------ */

jccTokenizer: PreinitObject
    tokenLookup = static new LookupTable(16, 32)
    identifiers = static new LookupTable(32, 128)
    jotaCodeFunctions = static new LookupTable(16, 32)

    execute()
    {
        foreach (local func in
            ['strcheck', 'execute',    'strlen',    'substr',   'lc',
             'lcfirst',  'move',       'toploc',    'tellroom', 'time',
             'null',     'type',       'index',     'rand',
             'ucfirst',  'substitute', 'setflag',   'location', 'uc',
             'setdest',  'name',       'shortname', 'log',      'ln',
             'event',    'killevent',  'statevent', 'tell',     'exp'])
            jotaCodeFunctions[func] = true;

        t3GetGlobalSymbols().forEachAssoc(new function(sym, val)
        {
            if (dataType(val) == TypeObject && val.ofKind(Token))
            {
                if (val.token) tokenLookup[val.token] = val;
                val.className = sym;
            }
        });
    }

    stringCleanUp1 = static new RexPattern('(%%c|;)<space>*<newline><space>*')
    stringCleanUp2 = static new RexPattern('<space>*<newline><space>*')
    wordMatch      = static new RexPattern('@?<Alpha|_><AlphaNum|_>*')
    percentMatch   = static new RexPattern('%%<AlphaNum|_|#|!>+')
    numberMatch    = static new RexPattern('(#(-)?)?[0-9]+(<period>[0-9]+)?([e|E](-)?[0-9]+)?')
    
    sourceLineNo = 1
    sourceLinePos = 0
    sourceReadMore = nil

    source = nil
    sourceIdx = 0
    sourceLen = 0
    
    tokens = nil
    tokenPositions = nil
    
    sourceAdvance()
    {
        sourceIdx++;
        while (sourceIdx > source.length()) sourceReadMore()();

        local ch = source.toUnicode(sourceIdx);
        if (ch == 10) { sourceLineNo++; sourceLinePos = 0; }
        else sourceLinePos++;
        return ch;
    }
    sourceAdvanceAmount(amount)
    {
        amount += sourceIdx;
        while (sourceIdx < amount) sourceAdvance();
    }
    
    tokenize()
    {
        local ch, chval;
        
        while (true)
        {
            try
            {
                chval = sourceAdvance();
            }
            catch (ETokenizerReachedEOF e)
            {
                addToken(tokEOF, nil);
                return;
            }
            if (sourceIdx > 64)
            {
                source = source.substr(sourceIdx);
                sourceIdx = 1;
            }
            
            if (chval <= 32) continue;

            local x = source.substr(sourceIdx, 2);
            switch (x)
            {
            case '//':
                while ((x = source.find('\n', sourceIdx)) == nil) sourceReadMore()();
                sourceIdx = x - 1; /* We don't have to update sourcepos - the \n will do that. */
                continue;
           
            case '/*':
                while ((x = source.find('*/', sourceIdx)) == nil) sourceReadMore()();
                // The final / is at (x + 1), so advance until we're at x - and then once more.
                while (sourceIdx <= x) sourceAdvance();
                continue;

            case '&&':
            case '||':
                sourceAdvance();
                addToken(tokOperator, x == '||' ? 'or' : 'and');
                continue;
            
            case '==':
            case '!=':
            case '>=':
            case '<=':
                sourceAdvance();
                
                addToken(tokOperator, x);
                continue;
            }
            
            ch = source.substr(sourceIdx, 1);
            if ('[](){}<>.,:;=!+-*/&?'.find(ch)) { addToken(tokOperator, ch); continue; }
            
            if (ch == '$')
            {
                while ((x = source.find('\n', sourceIdx)) == nil) sourceReadMore()();
                addToken(tokEmit, source.substr(sourceIdx + 1, x - sourceIdx - 1));
                sourceIdx = x - 1; /* We don't have to update sourcepos - the \n will do that. */
                continue;
            }
            
            if (ch is in ('"', '\''))
            {
                local pat = new RexPattern('(\\.|[^' + ch + '\\])*' + ch);
                while ((x = rexMatch(pat, source, sourceIdx + 1)) == nil) sourceReadMore()();

                addToken(tokString, source.substr(sourceIdx + 1, x - 1));

                sourceAdvanceAmount(x);

                continue;
            }
            
            x = rexMatch(wordMatch, source, sourceIdx);
            if (!x) x = rexMatch(percentMatch, source, sourceIdx);
            if (x) { addToken(tokWord, source.substr(sourceIdx, x)); sourceAdvanceAmount(x - 1); continue; }

            x = rexMatch(numberMatch, source, sourceIdx);
            if (x) { addToken(tokNumber, source.substr(sourceIdx, x)); sourceAdvanceAmount(x - 1); continue; }
            
            throw new ETokenizerError('Could not parse character ' + ch + ' in source');
        }
    }
    
    addToken(type, value)
    {
        local result, i, start, token;
        
        switch (type)
        {
        case tokString:
            value = rexReplace(stringCleanUp1, value, '%1', ReplaceAll);
            value = rexReplace(stringCleanUp2, value, ' ',  ReplaceAll);
            result = '';
            for (i = 1, local len = value.length(), start = 1; i <= len; ++i)
            {
                local qu = nil;
                if (value.toUnicode(i) == 92) // Backslash?
                {
                    switch (value.substr(i + 1, 1))
                    {
                        case '\\': qu = '\\'; break;
                        case '\'': qu = '\''; break;
                        case '\"': qu = '\"'; break;
                        case '\n': qu = '\n'; break;
                    }
                }
                if (qu)
                {
                    if (i != start) result += value.substr(start, i - start);
                    result += qu;
                    start = i + 2;
                    i++;
                }
                
            }
            if (i != start) result += value.substr(start, i - start);
            token = new StringToken(result);
            break;
        
        case tokNumber:
            if (value.substr(1, 1) == '#') value = value.substr(2);
            token = new NumberToken(toInteger(value));
            break;

        case tokOperator:
            token = tokenLookup[value];
            break;

        case tokWord:
            if (value.substr(1, 1) == '%')
            {
                if (value.length() >= 3) value = '%{' + value.substr(2, value.length() - 1) + '}';
                token = new StringToken(value);
                break;
            }
            
            value = value.toLower();
            if (value.substr(1, 1) == '@') { token = new JCFunctionToken(value.substr(2)); break; }
            if (jotaCodeFunctions.isKeyPresent(value)) { token = new JCFunctionToken(value); break; }

            token = tokenLookup[value];
            if (!token)
            {
                token = identifiers[value];
                if (!token) { identifiers[value] = token = new IdentifierToken(value); }
                token.token = value;
            }
            break;

        case tokEmit:
            token = new EmitToken(value);
            break;

        case tokEOF: 
            token = eofToken;
            break;
        }
        
        tokens.append(token);
        tokenPositions.append(sourceLineNo);
        tokenPositions.append(sourceLinePos);
    }
;

/* ------------------------------------------------------------------------ */
